/////////////////////////////////////////////////////////////////////////////
// ShellExt.cpp
// Implementation of the CShellExt class
//
// Wrapper class for Shell Functions
//
// All rights reserved.
//
// Written by Naveen Kohli (naveen@a3ds.com)
// Version 1.0
//
// Distribute freely, except: don't remove my name from the source or
// documentation (don't take credit for my work), mark your changes (don't
// get me blamed for your possible bugs), don't alter or remove this
// notice.
// No warrantee of any kind, express or implied, is included with this
// software; use at your own risk, responsibility for damages (if any) to
// anyone resulting from the use of this software rests entirely with the
// user.
//
// Send bug reports, bug fixes, enhancements, requests, flames, etc. to
//    naveen@a3ds.com
/////////////////////////////////////////////////////////////////////////////

#include "stdafx.h"
#include "NKShellAPI.h"
#include "ShellExt.h"

#ifdef _DEBUG
#undef THIS_FILE
static char THIS_FILE[]=__FILE__;
#define new DEBUG_NEW
#endif

//
//	Function:
//		CShellExt()
//
//	Parameters:
//
//	Returns:
//
//	Description:
//		CShellExt class constructor

CShellExt::CShellExt()
{
	m_bIsEnvInitialized = false;
	m_pMalloc = NULL;
	m_pDesktopFolder = NULL;
}

//
//	Function:
//		~CShellExt()
//
//	Parameters:
//
//	Returns:
//
//	Description:
//		CShellExt class Destructor
//

CShellExt::~CShellExt()
{
	m_pMalloc->Release ();
	m_pDesktopFolder->Release ();
}

//
//	Function:
//		HRESULT IsInitialized () const
//
//	Parameters:
//
//	Returns:
//		S_OK	Success
//		E_FAIL	Failure
//
//	Description:
//		Call this function to check if the environment interface pointers have been
//		initialized or not.
//

HRESULT
CShellExt::IsInitialized () const
{
	HRESULT hr = (m_bIsEnvInitialized && m_pMalloc != NULL && m_pDesktopFolder != NULL) 
		? S_OK : E_FAIL;
	return (hr);
}

//
//	Function:
//		const FolderItemList *GetCompsOnNetwork ()
//
//	Parameters:
//
//	Returns:
//		List of folder items
//
//	Description:
//		Call this function to get the list of folder items corresponding to each user
//		machine logged on to network. In case of any error, function returns NULL.
//

const FolderItemList *
CShellExt::GetCompsOnNetwork ()
{
	HRESULT hr = GetNetworkComps ();
	if (FAILED (hr)) {
		return (NULL);
	}
	return &m_ItemList;
}

//
//	Function:
//		const FolderItemList *GetPrintersOnNetwork ()
//
//	Parameters:
//
//	Returns:
//		List of folder items
//
//	Description:
//		Call this function to get the list of folder items corresponding to each printer
//		connected to network or local machine. In case of any error, function returns NULL.
//

const FolderItemList *
CShellExt::GetPrintersOnNetwork ()
{
	HRESULT hr = GetNetworkPrinters ();
	if (FAILED (hr)) {
		return (NULL);
	}
	return &m_ItemList;
}

//
//	Function:
//		const FolderItemList *GetRecycleBinItems ()
//
//	Parameters:
//
//	Returns:
//		List of folder items
//
//	Description:
//		Call this function to get the list of folder items corresponding to each item in
//		the recycle bin of local machine. In case of any error, function returns NULL.
//

const FolderItemList *
CShellExt::GetRecycleBinItems ()
{
	HRESULT hr = GetRecycleBin ();
	if (FAILED (hr)) {
		return (NULL);
	}
	return &m_ItemList;
}

//
//	Function:
//		HRESULT GetNetworkComps ()
//
//	Parameters:
//
//	Returns:
//		S_OK/E_FAIL		Success/Failure
//
//	Description:
//		This helper function gets called internally to get the list of all user machines
//		logged on to network.
//		

HRESULT
CShellExt::GetNetworkComps ()
{
	HRESULT hr = S_OK;
	ULONG celtFetched;
	STRRET strDispName;
	LPITEMIDLIST pidlItems = NULL;
	IShellFolder *pNethoodFolder = NULL;
	IEnumIDList *pEnumIds = NULL;

	// Initialize the environment variables.

	if ((FAILED (hr = InitializeEnvironment ()))) {
		return (hr);
	}

	// Get the folder corresponding to network neighbourhood.

	hr = GetNode (CSIDL_NETWORK, (void **)&pNethoodFolder);

	// Enumerate the objects under this folder.

	hr = pNethoodFolder->EnumObjects (NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS,
		&pEnumIds);
	if (FAILED (hr)) {
		goto FreeInterfaces;
	}

	// TODO: Encapsulate this enumeration process in a recursive function, to be used
	//		 by each query function.

	// Check each of the objects in this enumeration.

	while ((hr = pEnumIds->Next (1, &pidlItems, &celtFetched)) != S_FALSE && celtFetched == 1) {
		hr = pNethoodFolder->GetDisplayNameOf (pidlItems, SHGDN_INFOLDER, &strDispName);
		if (FAILED (hr)) {
			goto FreeInterfaces;
		}
		
		FOLDERITEM item;
		HICON hIconLarge = NULL;
		HICON hIconSmall = NULL;
		LPSTR name;

		// Get the icon associated with this folder.

		hr = GetIconsForFolderNode (pNethoodFolder, pidlItems, hIconSmall, hIconLarge);
		if (SUCCEEDED (hr)) {
			item.hIconSmall = hIconSmall;
			item.hIconLarge = hIconLarge;
		}
		
		// Convert the display name to CString.

		StrRetToBuf (pidlItems, &strDispName, name);

		item.strFolderName = name;
		m_ItemList.AddHead (item);
	}
	
FreeInterfaces:

	if (pNethoodFolder != NULL) {
		pNethoodFolder->Release ();
	}
	
	return (hr);
}

//
//	Function:
//		HRESULT GetNetworkPrinters ()
//
//	Parameters:
//
//	Returns:
//		S_OK/E_FAIL		Success/Failure
//
//	Description:
//		This helper function gets called internally to get the list of all printers
//		installed on network and local machine.
//		

HRESULT
CShellExt::GetNetworkPrinters ()
{
	HRESULT hr = S_OK;
	ULONG celtFetched;
	STRRET strDispName;
	LPITEMIDLIST pidlItems = NULL;
	IShellFolder *pPrinterFolder = NULL;
	IEnumIDList *pEnumIds = NULL;

	// Initialize the environment variables.

	if ((FAILED (hr = InitializeEnvironment ()))) {
		return (hr);
	}

	// Get the folder corresponding to printers installed.

	hr = GetNode (CSIDL_PRINTERS, (void **)&pPrinterFolder);

	// TODO: Encapsulate this enumeration process in a recursive function, to be used
	//		 by each query function.

	// Enumerate the objects under this folder.

	hr = pPrinterFolder->EnumObjects (NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS,
		&pEnumIds);
	if (FAILED (hr)) {
		goto FreeInterfaces;
	}

	// Check each of the objects in this enumeration.

	while ((hr = pEnumIds->Next (1, &pidlItems, &celtFetched)) != S_FALSE && celtFetched == 1) {
		hr = pPrinterFolder->GetDisplayNameOf (pidlItems, SHGDN_INFOLDER, &strDispName);
		if (FAILED (hr)) {
			goto FreeInterfaces;
		}
		
		FOLDERITEM item;
		HICON hIconLarge = NULL;
		HICON hIconSmall = NULL;
		LPSTR name;

		// Get the icon associated with this folder.

		hr = GetIconsForFolderNode (pPrinterFolder, pidlItems, hIconSmall, hIconLarge);
		if (SUCCEEDED (hr)) {
			item.hIconSmall = hIconSmall;
			item.hIconLarge = hIconLarge;
		}
		
		StrRetToBuf (pidlItems, &strDispName, name);
		item.strFolderName = name;
		m_ItemList.AddHead (item);
	}
	
FreeInterfaces:

	if (pPrinterFolder != NULL) {
		pPrinterFolder->Release ();
	}
	return (hr);
}

//
//	Function:
//		HRESULT GetRecycleBin ()
//
//	Parameters:
//
//	Returns:
//		S_OK/E_FAIL		Success/Failure
//
//	Description:
//		This helper function gets called internally to get the list of all items present
//		in the recycle bin of local machine.
//		

HRESULT
CShellExt::GetRecycleBin ()
{
	HRESULT hr = S_OK;
	ULONG celtFetched;
	STRRET strDispName;
	LPITEMIDLIST pidlItems = NULL;
	IShellFolder *pRecycleFolder = NULL;
	IExtractIcon *pExtractIcon = NULL;
	IEnumIDList *pEnumIds = NULL;

	// Initialize the environment variables.

	if ((FAILED (hr = InitializeEnvironment ()))) {
		return (hr);
	}

	// Get the folder corresponding to recycle bin.

	hr = GetNode (CSIDL_BITBUCKET, (void **)&pRecycleFolder);

	// Enumerate the objects under this folder.

	hr = pRecycleFolder->EnumObjects (NULL, SHCONTF_FOLDERS | SHCONTF_NONFOLDERS,
		&pEnumIds);
	if (FAILED (hr)) {
		goto FreeInterfaces;
	}

	// Check each of the objects in this enumeration.

	while ((hr = pEnumIds->Next (1, &pidlItems, &celtFetched)) != S_FALSE && celtFetched == 1) {
		hr = pRecycleFolder->GetDisplayNameOf (pidlItems, SHGDN_INFOLDER, &strDispName);
		if (FAILED (hr)) {
			goto FreeInterfaces;
		}

		FOLDERITEM item;
		HICON hIconLarge = NULL;
		HICON hIconSmall = NULL;
		LPSTR name;

		// Get the icon associated with this folder.

		hr = GetIconsForFolderNode (pRecycleFolder, pidlItems, hIconSmall, hIconLarge);
		if (SUCCEEDED (hr)) {
			item.hIconSmall = hIconSmall;
			item.hIconLarge = hIconLarge;
		}
		
		StrRetToBuf (pidlItems, &strDispName, name);
		item.strFolderName = CString (name);
		m_ItemList.AddHead (item);
	}
	
FreeInterfaces:

	if (pRecycleFolder != NULL) {
		pRecycleFolder->Release ();
	}
	
	return (hr);
}

//
//	Function:
//		HRESULT InitializeEnvironment ()
//
//	Parameters:
//
//	Returns:
//		S_OK/E_FAIL		Success/Failure
//
//	Description:
//		Call this function to initialize the IMalloc and IShellFolder interface
//		pointers for the desktop folder. Function makes sure that if these variables
//		have already been initialized, then it won't do it again.
//		

HRESULT
CShellExt::InitializeEnvironment ()
{
	HRESULT hr = S_OK;

	if (!FAILED (hr = IsInitialized ())) {
		return S_OK;
	}

	// If the IMalloc interface has not been obtained, get it.

	if (m_pMalloc == NULL) {
		hr = SHGetMalloc (&m_pMalloc);
		if (FAILED (hr)) {
			return (hr);
		}
	}

	// Get IShellFolder interface pointer for Desktop folder.

	if (m_pDesktopFolder == NULL) {
		hr = SHGetDesktopFolder (&m_pDesktopFolder);
		if (FAILED (hr)) {
			return (hr);
		}
	}

	m_bIsEnvInitialized = true;
	return (true);
}

//
//	Function:
//		HRESULT GetNode (int folderNode, void **pFolderNode)
//
//	Parameters:
//		folderNode	Type of folder requested requested
//		pFolderNode	Pointer to shell folder interface
//
//	Returns:
//		S_OK/E_FAIL	Success/Failure
//
//	Description:
//		This function is used internally to bind the requested folder type's IShellFolder
//		interface.
//		

HRESULT
CShellExt::GetNode (int folderNode, void **pFolderNode)
{
	HRESULT hr = S_OK;
	LPITEMIDLIST netItemIdLst = NULL;

	// Get the folder corresponding to network neighbourhood.

	hr = SHGetSpecialFolderLocation (NULL, folderNode, &netItemIdLst);
	if (FAILED (hr)) {
		return (hr);
	}

	// Get the interface pointer to sub folder of network.

	hr = m_pDesktopFolder->BindToObject (netItemIdLst, NULL, IID_IShellFolder,
		pFolderNode);
	
	return (hr);
}

//
//	Function:
//		HRESULT GetIconsForFolderNode (IShellFolder *pFolderNode, LPITEMIDLIST pidlItems,
//									  HICON &hIconSm, HICON &hIconLg)
//
//	Parameters:
//		pFolderNode		Pointer to IShellFolder interface
//		pidlItems		List of SHITMIDs in the folder
//		hIconSm			Reference to handle to small icon for folder
//		hIconLg			Reference to handle to large icon for folder
//
//	Returns:
//		S_OK/E_FAIL		Success/Failure
//
//	Description:
//		Function is used internally to get handles to large and small icon corresponding
//		to a given IShellFolder interface. First it gets the IExtractIcon interface for
//		folder interface. If the folder has impleemnted this interface, then the handles
//		to icons are extracted from IExtractIcon interface.
//		

HRESULT
CShellExt::GetIconsForFolderNode (IShellFolder *pFolderNode, LPITEMIDLIST pidlItems,
								  HICON &hIconSm, HICON &hIconLg)
{
	HRESULT hr = S_OK;
	IExtractIcon *pExtractIcon = NULL;

	ASSERT (pFolderNode != NULL);

	// Get the IExtractIcon interface for the given IShellFolder interface.

	hr = pFolderNode->GetUIObjectOf (NULL, 1, const_cast<LPCITEMIDLIST *>(&pidlItems),
		IID_IExtractIcon, NULL, reinterpret_cast<LPVOID *>(&pExtractIcon));
	
	if (SUCCEEDED (hr)) {
		TCHAR str[MAX_PATH] = {0};
		int index;
		UINT flags;

		// Get the file location where the icons are stored.

		hr = pExtractIcon->GetIconLocation (0, str, MAX_PATH, &index, &flags);
		if (SUCCEEDED (hr)) {
			HICON hiconLarge = NULL;
			HICON hiconSmall = NULL;
			UINT nIconSize = MAKELONG (32, 16);
			LPCSTR pszFile = (LPCSTR)str;

			// Extract the icon handles from the location.

			hr = pExtractIcon->Extract (str, index, &hIconLg, &hIconSm, nIconSize);
			pExtractIcon->Release ();
		}
	}

	return (hr);
}

//
//	Function:
//		void StrRetToBuf (LPITEMIDLIST pidl, LPSTRRET lpStrt, LPSTR &lpStr)
//
//	Parameters:
//
//	Returns:
//
//	Description:
//		This function has been introduced in the Windows2000 SDK. Fo rnow it is being
//		used to LPSTRRET data to LPSTR type data.
//		

void
CShellExt::StrRetToBuf (LPITEMIDLIST pidl, LPSTRRET lpStrt, LPSTR &lpStr)
{
	int cch;

	switch (lpStrt->uType) {
		case STRRET_WSTR:
			cch = WideCharToMultiByte (CP_OEMCP, NULL, lpStrt->pOleStr, -1,
				NULL, 0, NULL, NULL);

			// Allocate the memeory for copying string.

			if ((lpStr = (LPSTR)m_pMalloc->Alloc (cch)) != NULL) {
				WideCharToMultiByte (CP_OEMCP, NULL, lpStrt->pOleStr, -1,
					lpStr, cch, NULL, NULL);
			}
			break;
		case STRRET_OFFSET:
			lpStr = ((char *)pidl) + lpStrt->uOffset;
			break;
		case STRRET_CSTR:
			lpStr = lpStrt->cStr;
			break;
	}
}
